from ..schema import get_current_schema


def make_key(key, key_prefix, version):
    """
    Tenant aware function to generate a cache key.

    Constructs the key used by all other methods. Prepends the tenant
    `schema_name` and `key_prefix'.
    """
    current_schema = get_current_schema()
    return "%s:%s:%s:%s" % (current_schema.schema_name, key_prefix, version, key)


def reverse_key(key):
    """
    Tenant aware function to reverse a cache key.

    Required for django-redis REVERSE_KEY_FUNCTION setting.
    """
    return key.split(":", 3)[3]



from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db.models import CharField, Q, Value as V
from django.db.models.functions import Concat
from django.db.utils import ProgrammingError

from ...schema import get_current_schema
from ...utils import create_schema, dynamic_models_exist, get_clone_reference, get_tenant_model
from ._executors import parallel, sequential

EXECUTORS = {"sequential": sequential, "parallel": parallel}


class WrappedSchemaOption:
    scope = "all"
    specific_schemas = None

    allow_interactive = True
    allow_wildcards = True

    def add_arguments(self, parser):
        if self.allow_interactive:
            parser.add_argument(
                "--noinput",
                "--no-input",
                action="store_false",
                dest="interactive",
                help="Tells Django to NOT prompt the user for input of any kind.",
            )
        parser.add_argument(
            "-s", "--schema", nargs="+", dest="schemas", help="Schema(s) to execute the current command"
        )
        parser.add_argument(
            "-as",
            "--include-all-schemas",
            action="store_true",
            dest="all_schemas",
            help="Include all schemas when executing the current command",
        )
        parser.add_argument(
            "-ss",
            "--include-static-schemas",
            action="store_true",
            dest="static_schemas",
            help="Include all static schemas when executing the current command",
        )
        parser.add_argument(
            "-ds",
            "--include-dynamic-schemas",
            action="store_true",
            dest="dynamic_schemas",
            help="Include all dynamic schemas when executing the current command",
        )
        parser.add_argument(
            "-ts",
            "--include-tenant-schemas",
            action="store_true",
            dest="tenant_schemas",
            help="Include all tenant-like schemas when executing the current command",
        )
        parser.add_argument(
            "-x",
            "--exclude-schema",
            nargs="+",
            dest="excluded_schemas",
            help="Schema(s) to exclude when executing the current command",
        )
        parser.add_argument(
            "--parallel",
            dest="parallel",
            action="store_true",
            help="Run command in parallel mode",
        )
        parser.add_argument(
            "--no-create-schemas",
            dest="skip_schema_creation",
            action="store_true",
            help="Skip automatic creation of non-existing schemas",
        )

    def get_schemas_from_options(self, **options):
        skip_schema_creation = options.get("skip_schema_creation", False)
        try:
            schemas = self._get_schemas_from_options(**options)
        except ProgrammingError:
            # This happens with unmigrated database.
            # It can also happen when the tenant model contains unapplied migrations that break.
            raise CommandError(
                "Error while attempting to retrieve dynamic schemas. "
                "Perhaps you need to migrate the 'public' schema first?"
            )
        if self.specific_schemas is not None:
            schemas = [x for x in schemas if x in self.specific_schemas]
            if not schemas:
                raise CommandError("This command can only run in %s" % self.specific_schemas)
        if not skip_schema_creation:
            for schema in schemas:
                create_schema(schema, check_if_exists=True, sync_schema=False, verbosity=0)
        return schemas

    def get_executor_from_options(self, **options):
        return EXECUTORS["parallel"] if options.get("parallel") else EXECUTORS["sequential"]

    def get_scope_display(self):
        return "|".join(self.specific_schemas or []) or self.scope

    def _get_schemas_from_options(self, **options):
        schemas = options.get("schemas") or []
        excluded_schemas = options.get("excluded_schemas") or []
        include_all_schemas = options.get("all_schemas") or False
        include_static_schemas = options.get("static_schemas") or False
        include_dynamic_schemas = options.get("dynamic_schemas") or False
        include_tenant_schemas = options.get("tenant_schemas") or False
        dynamic_ready = dynamic_models_exist()
        allow_static = self.scope in ["all", "static"]
        allow_dynamic = self.scope in ["all", "dynamic"]
        clone_reference = get_clone_reference()

        if (
            not schemas
            and not include_all_schemas
            and not include_static_schemas
            and not include_dynamic_schemas
            and not include_tenant_schemas
        ):
            if not self.allow_interactive:
                include_all_schemas = True
            elif options.get("interactive", True):
                schema = input(
                    "Enter schema to run command (leave blank for running on '%s' schemas): " % self.get_scope_display()
                ).strip()

                if schema:
                    schemas.append(schema)
                else:
                    include_all_schemas = True
            else:
                raise CommandError("No schema provided")

        TenantModel = get_tenant_model()
        static_schemas = [x for x in settings.TENANTS.keys() if x != "default"] if allow_static else []
        dynamic_schemas = (
            TenantModel.objects.values_list("schema_name", flat=True) if dynamic_ready and allow_dynamic else []
        )
        if clone_reference and allow_static:
            static_schemas.append(clone_reference)

        schemas_to_return = set()

        if include_all_schemas:
            if not self.allow_wildcards or (not allow_static and not allow_dynamic):
                raise CommandError("Including all schemas is NOT allowed")
            schemas_to_return = schemas_to_return.union(static_schemas + list(dynamic_schemas))
        if include_static_schemas:
            if not self.allow_wildcards or not allow_static:
                raise CommandError("Including static schemas is NOT allowed")
            schemas_to_return = schemas_to_return.union(static_schemas)
        if include_dynamic_schemas:
            if not self.allow_wildcards or not allow_dynamic:
                raise CommandError("Including dynamic schemas is NOT allowed")
            schemas_to_return = schemas_to_return.union(dynamic_schemas)
        if include_tenant_schemas:
            if not self.allow_wildcards or not allow_dynamic:
                raise CommandError("Including tenant-like schemas is NOT allowed")
            schemas_to_return = schemas_to_return.union(dynamic_schemas)
            if clone_reference:
                schemas_to_return.add(clone_reference)

        for schema in schemas:
            if schema in settings.TENANTS and schema != "default" and allow_static:
                schemas_to_return.add(schema)
            elif schema == clone_reference:
                schemas_to_return.add(schema)
            elif dynamic_ready and TenantModel.objects.filter(schema_name=schema).exists() and allow_dynamic:
                schemas_to_return.add(schema)

        schemas = list(set(schemas) - schemas_to_return)

        for schema in schemas:
            local = []
            if allow_static:
                local += [
                    schema_name
                    for schema_name, data in settings.TENANTS.items()
                    if schema_name not in ["public", "default"]
                    and any(x for x in data["DOMAINS"] if x.startswith(schema))
                ]
            if dynamic_ready and allow_dynamic:
                local += (
                    TenantModel.objects.annotate(
                        route=Concat("domains__domain", V("/"), "domains__folder", output_field=CharField())
                    )
                    .filter(Q(schema_name=schema) | Q(domains__domain__istartswith=schema) | Q(route=schema))
                    .distinct()
                    .values_list("schema_name", flat=True)
                )
            if not local:
                raise CommandError("No schema found for '%s'" % schema)
            if len(local) > 1:
                raise CommandError(
                    "More than one tenant found for schema '%s' by domain, please, narrow down the filter" % schema
                )
            schemas_to_return.add(local.pop())

        excluded = []
        for schema in excluded_schemas:
            local = []
            if schema in ["public", clone_reference]:
                excluded.append(schema)
                continue
            local += [
                schema_name
                for schema_name, data in settings.TENANTS.items()
                if schema_name not in ["public", "default", clone_reference]
                and any(x for x in data["DOMAINS"] if x.startswith(schema))
            ]
            local += (
                TenantModel.objects.annotate(
                    route=Concat("domains__domain", V("/"), "domains__folder", output_field=CharField())
                )
                .filter(Q(schema_name=schema) | Q(domains__domain__istartswith=schema) | Q(route=schema))
                .distinct()
                .values_list("schema_name", flat=True)
            )
            if not local:
                raise CommandError("No schema found for '%s' (excluded)" % schema)
            if len(local) > 1:
                raise CommandError(
                    "More than one tenant found for schema '%s' by domain (excluded), please, narrow down the filter"
                    % schema
                )
            excluded += local
        schemas_to_return -= set(excluded)

        return (
            list(schemas_to_return)
            if "public" not in schemas_to_return
            else ["public"] + list(schemas_to_return - {"public"})
        )


class TenantCommand(WrappedSchemaOption, BaseCommand):
    def handle(self, *args, **options):
        schemas = self.get_schemas_from_options(**options)
        executor = self.get_executor_from_options(**options)
        executor(schemas, self, "_raw_handle_tenant", args, options, pass_schema_in_kwargs=True)

    def _raw_handle_tenant(self, *args, **kwargs):
        kwargs.pop("schema_name")
        self.handle_tenant(get_current_schema(), *args, **kwargs)

    def handle_tenant(self, tenant, *args, **options):
        pass


class StaticTenantCommand(TenantCommand):
    scope = "static"


class DynamicTenantCommand(TenantCommand):
    scope = "dynamic"



import re

from django.conf import settings
from django.http import Http404
from django.shortcuts import redirect
from django.urls import clear_url_caches, set_urlconf

from .schema import SchemaDescriptor, activate, activate_public
from .urlresolvers import get_urlconf_from_schema
from .utils import get_domain_model, remove_www


class TenantMiddleware:
    """
    This middleware should be placed at the very top of the middleware stack.
    Selects the proper static/dynamic schema using the request host. Can fail in
    various ways which is better than corrupting or revealing data.
    """

    TENANT_NOT_FOUND_EXCEPTION = Http404

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        hostname = remove_www(request.get_host().split(":")[0])

        activate_public()
        tenant = None

        # Checking for static tenants
        for schema, data in settings.TENANTS.items():
            if schema in ["public", "default"]:
                continue
            if hostname in data["DOMAINS"]:
                tenant = SchemaDescriptor.create(schema_name=schema, domain_url=hostname)
                break

        # Checking for dynamic tenants
        else:
            DomainModel = get_domain_model()
            prefix = request.path.split("/")[1]
            try:
                domain = DomainModel.objects.select_related("tenant").get(domain=hostname, folder=prefix)
            except DomainModel.DoesNotExist:
                try:
                    domain = DomainModel.objects.select_related("tenant").get(domain=hostname, folder="")
                except DomainModel.DoesNotExist:
                    domain = None
            if domain:
                tenant = domain.tenant
                tenant.domain_url = hostname
                tenant.folder = None
                request.strip_tenant_from_path = lambda x: x
                if prefix and domain.folder == prefix:
                    tenant.folder = prefix
                    request.strip_tenant_from_path = lambda x: re.sub(r"^/{}/".format(prefix), "/", x)
                    clear_url_caches()  # Required to remove previous tenant prefix from cache (#8)
                if domain.redirect_to_primary:
                    primary_domain = tenant.domains.get(is_primary=True)
                    path = request.strip_tenant_from_path(request.path)
                    return redirect(primary_domain.absolute_url(path), permanent=True)

        # Checking fallback domains
        if not tenant:
            for schema, data in settings.TENANTS.items():
                if schema in ["public", "default"]:
                    continue
                if hostname in data.get("FALLBACK_DOMAINS", []):
                    tenant = SchemaDescriptor.create(schema_name=schema, domain_url=hostname)
                    break

        # No tenant found from domain / folder
        if not tenant:
            raise self.TENANT_NOT_FOUND_EXCEPTION("No tenant for hostname '%s'" % hostname)

        request.tenant = tenant
        urlconf = get_urlconf_from_schema(tenant)
        request.urlconf = urlconf
        set_urlconf(urlconf)

        activate(tenant)
        return self.get_response(request)



from django.db.models.indexes import Index


def get_constraints(self, cursor, table_name):
    """
    Retrieve any constraints or keys (unique, pk, fk, check, index) across
    one or more columns. Also retrieve the definition of expression-based
    indexes.
    """
    constraints = {}
    # Loop over the key table, collecting things as constraints. The column
    # array must return column names in the same order in which they were
    # created.
    # The subquery containing generate_series can be replaced with
    # "WITH ORDINALITY" when support for PostgreSQL 9.3 is dropped.
    cursor.execute(
        """
        SELECT
            c.conname,
            array(
                SELECT attname
                FROM (
                    SELECT unnest(c.conkey) AS colid,
                           generate_series(1, array_length(c.conkey, 1)) AS arridx
                ) AS cols
                JOIN pg_attribute AS ca ON cols.colid = ca.attnum
                WHERE ca.attrelid = c.conrelid
                ORDER BY cols.arridx
            ),
            c.contype,
            (SELECT fkc.relname || '.' || fka.attname
            FROM pg_attribute AS fka
            JOIN pg_class AS fkc ON fka.attrelid = fkc.oid
            WHERE fka.attrelid = c.confrelid AND fka.attnum = c.confkey[1]),
            cl.reloptions
        FROM pg_constraint AS c
        JOIN pg_class AS cl ON c.conrelid = cl.oid
        JOIN pg_namespace AS ns ON cl.relnamespace = ns.oid
        WHERE ns.nspname = %s AND cl.relname = %s
    """,
        [self.connection._schema.schema_name, table_name],
    )
    for constraint, columns, kind, used_cols, options in cursor.fetchall():
        constraints[constraint] = {
            "columns": columns,
            "primary_key": kind == "p",
            "unique": kind in ["p", "u"],
            "foreign_key": tuple(used_cols.split(".", 1)) if kind == "f" else None,
            "check": kind == "c",
            "index": False,
            "definition": None,
            "options": options,
        }
    # Now get indexes
    # The row_number() function for ordering the index fields can be
    # replaced by WITH ORDINALITY in the unnest() functions when support
    # for PostgreSQL 9.3 is dropped.
    cursor.execute(
        """
        SELECT
            indexname, array_agg(attname ORDER BY rnum), indisunique, indisprimary,
            array_agg(ordering ORDER BY rnum), amname, exprdef, s2.attoptions
        FROM (
            SELECT
                row_number() OVER () as rnum, c2.relname as indexname,
                idx.*, attr.attname, am.amname,
                CASE
                    WHEN idx.indexprs IS NOT NULL THEN
                        pg_get_indexdef(idx.indexrelid)
                END AS exprdef,
                CASE am.amname
                    WHEN 'btree' THEN
                        CASE (option & 1)
                            WHEN 1 THEN 'DESC' ELSE 'ASC'
                        END
                END as ordering,
                c2.reloptions as attoptions
            FROM (
                SELECT
                    *, unnest(i.indkey) as key, unnest(i.indoption) as option
                FROM pg_index i
            ) idx
            LEFT JOIN pg_class c ON idx.indrelid = c.oid
            LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
            LEFT JOIN pg_class c2 ON idx.indexrelid = c2.oid
            LEFT JOIN pg_am am ON c2.relam = am.oid
            LEFT JOIN pg_attribute attr ON attr.attrelid = c.oid AND attr.attnum = idx.key
            WHERE c.relname = %s and n.nspname = %s
        ) s2
        GROUP BY indexname, indisunique, indisprimary, amname, exprdef, attoptions;
    """,
        [table_name, self.connection._schema.schema_name],
    )
    for index, columns, unique, primary, orders, type_, definition, options in cursor.fetchall():
        if index not in constraints:
            constraints[index] = {
                "columns": columns if columns != [None] else [],
                "orders": orders if orders != [None] else [],
                "primary_key": primary,
                "unique": unique,
                "foreign_key": None,
                "check": False,
                "index": True,
                "type": Index.suffix if type_ == "btree" else type_,
                "definition": definition,
                "options": options,
            }
    return constraints



from django.conf import settings
from django.core.management import call_command
from django.test import TestCase

from ..schema import SchemaDescriptor, activate, activate_public
from ..utils import get_clone_reference, get_domain_model, get_tenant_model

ALLOWED_TEST_DOMAIN = ".test.com"


class BaseTenantTestCaseMixin:
    @classmethod
    def get_verbosity(cls):
        return 0

    @classmethod
    def add_allowed_test_domain(cls):
        cls.BACKUP_ALLOWED_HOSTS = settings.ALLOWED_HOSTS
        # ALLOWED_HOSTS is a special setting of Django setup_test_environment so we can't modify it with helpers
        if ALLOWED_TEST_DOMAIN not in settings.ALLOWED_HOSTS:
            settings.ALLOWED_HOSTS += [ALLOWED_TEST_DOMAIN]

    @classmethod
    def remove_allowed_test_domain(cls):
        settings.ALLOWED_HOSTS = cls.BACKUP_ALLOWED_HOSTS

    @classmethod
    def sync_public(cls):
        call_command("migrateschema", schemas=["public"], verbosity=0)


class StaticTenantTestCase(BaseTenantTestCaseMixin, TestCase):
    schema_name = None  # Meant to be set by subclasses
    tenant = None

    @classmethod
    def setUpClass(cls):
        assert (
            cls.schema_name in settings.TENANTS
        ), f"{cls.__name__}.schema_name must be defined to a valid static tenant"
        assert (
            cls.schema_name not in ["public", "default"] and cls.schema_name != get_clone_reference()
        ), f"{cls.__name__}.schema_name must be defined to a valid static tenant"
        super(TestCase, cls).setUpClass()
        cls.sync_public()
        cls.add_allowed_test_domain()
        domain = (
            settings.TENANTS[cls.schema_name]["DOMAINS"][0]
            if settings.TENANTS[cls.schema_name]["DOMAINS"]
            else cls.schema_name + ALLOWED_TEST_DOMAIN
        )
        cls.tenant = SchemaDescriptor.create(schema_name=cls.schema_name, domain_url=domain)
        activate(cls.tenant)
        cls.cls_atomics = cls._enter_atomics()
        try:
            cls.setUpTestData()
        except Exception:
            cls._rollback_atomics(cls.cls_atomics)
            raise

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        activate_public()
        cls.remove_allowed_test_domain()


class DynamicTenantTestCase(BaseTenantTestCaseMixin, TestCase):
    tenant = None
    domain = None

    @classmethod
    def setup_tenant(cls, tenant):
        """
        Add any additional setting to the tenant before it get saved. This is required if you have
        required fields.
        :param tenant:
        :return:
        """
        pass

    @classmethod
    def setup_domain(cls, domain):
        """
        Add any additional setting to the domain before it get saved. This is required if you have
        required fields.
        :param domain:
        :return:
        """
        pass

    @classmethod
    def setUpClass(cls):
        super(TestCase, cls).setUpClass()
        cls.sync_public()
        cls.add_allowed_test_domain()
        cls.tenant = get_tenant_model()(schema_name=cls.get_test_schema_name())
        cls.setup_tenant(cls.tenant)
        cls.tenant.save(verbosity=cls.get_verbosity())
        tenant_domain = cls.get_test_tenant_domain()
        cls.domain = get_domain_model()(tenant=cls.tenant, domain=tenant_domain)
        cls.setup_domain(cls.domain)
        cls.domain.save()
        activate(cls.tenant)
        cls.cls_atomics = cls._enter_atomics()
        try:
            cls.setUpTestData()
        except Exception:
            cls._rollback_atomics(cls.cls_atomics)
            raise

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        activate_public()
        cls.domain.delete()
        cls.tenant.delete(force_drop=True)
        cls.remove_allowed_test_domain()

    @classmethod
    def get_test_tenant_domain(cls):
        return "tenant.test.com"

    @classmethod
    def get_test_schema_name(cls):
        return "test"


class TenantTestCase(DynamicTenantTestCase):
    pass


class FastDynamicTenantTestCase(DynamicTenantTestCase):
    @classmethod
    def flush_data(cls):
        """
        Do you want to flush the data out of the tenant database.
        :return: bool
        """
        return True

    @classmethod
    def use_existing_tenant(cls):
        """
        Gets called if a existing tenant is found in the database
        """
        pass

    @classmethod
    def use_new_tenant(cls):
        """
        Gets called if a new tenant is created in the database
        """
        pass

    @classmethod
    def setup_test_tenant_and_domain(cls):
        cls.tenant = get_tenant_model()(schema_name=cls.get_test_schema_name())
        cls.setup_tenant(cls.tenant)
        cls.tenant.save(verbosity=cls.get_verbosity())

        # Set up domain
        tenant_domain = cls.get_test_tenant_domain()
        cls.domain = get_domain_model()(tenant=cls.tenant, domain=tenant_domain)
        cls.setup_domain(cls.domain)
        cls.domain.save()
        cls.use_new_tenant()

    @classmethod
    def setUpClass(cls):
        TenantModel = get_tenant_model()
        test_schema_name = cls.get_test_schema_name()
        if TenantModel.objects.filter(schema_name=test_schema_name).exists():
            cls.tenant = TenantModel.objects.filter(schema_name=test_schema_name).first()
            cls.use_existing_tenant()
        else:
            cls.setup_test_tenant_and_domain()

        activate(cls.tenant)

    @classmethod
    def tearDownClass(cls):
        TenantModel = get_tenant_model()
        test_schema_name = cls.get_test_schema_name()
        TenantModel.objects.filter(schema_name=test_schema_name).delete()
        activate_public()

    def _fixture_teardown(self):
        if self.flush_data():
            super()._fixture_teardown()


class FastTenantTestCase(FastDynamicTenantTestCase):
    pass



import psycopg2
from django.core.exceptions import ImproperlyConfigured
from django.db.utils import DatabaseError

from ..schema import SchemaDescriptor
from ..utils import check_schema_name, get_limit_set_calls
from .introspection import DatabaseSchemaIntrospection
from .settings import EXTRA_SEARCH_PATHS, original_backend

IntegrityError = psycopg2.IntegrityError


class DatabaseWrapper(original_backend.DatabaseWrapper):
    def __init__(self, *args, **kwargs):
        self._schema = None
        self._search_path_set = None
        super().__init__(*args, **kwargs)

        # Use a patched version of the DatabaseIntrospection that only returns the table list for the
        # currently selected schema.
        self.introspection = DatabaseSchemaIntrospection(self)
        self._set_schema_to_public()

    def close(self):
        self._search_path_set = False
        super().close()

    def _set_schema(self, schema_descriptor):
        assert isinstance(
            schema_descriptor, SchemaDescriptor
        ), "'set_schema' must be called with a SchemaDescriptor descendant"
        self._schema = schema_descriptor
        self._search_path_set = False

    def _set_schema_to_public(self):
        self._set_schema(SchemaDescriptor.create("public"))

    def _get_search_path(self):
        search_path = ["public"] if self._schema.schema_name == "public" else [self._schema.schema_name, "public"]
        search_path.extend(EXTRA_SEARCH_PATHS)
        return search_path

    def _cursor(self, name=None):
        if name:
            # Only supported and required by Django 1.11 (server-side cursor)
            cursor = super()._cursor(name=name)
        else:
            cursor = super()._cursor()

        # optionally limit the number of executions - under load, the execution
        # of `set search_path` can be quite time consuming
        if (not get_limit_set_calls()) or not self._search_path_set:
            # Actual search_path modification for the cursor. Database will
            # search schemas from left to right when looking for the object
            # (table, index, sequence, etc.).
            if not self._schema:
                raise ImproperlyConfigured("Database schema not set. Did you forget to call set_schema()?")

            check_schema_name(self._schema.schema_name)
            search_path = ",".join(self._get_search_path())

            if name:
                # Named cursor can only be used once
                cursor_for_search_path = self.connection.cursor()
            else:
                # Reuse
                cursor_for_search_path = cursor

            # In the event that an error already happened in this transaction and we are going
            # to rollback we should just ignore database error when setting the search_path
            # if the next instruction is not a rollback it will just fail also, so
            # we do not have to worry that it's not the good one
            try:
                cursor_for_search_path.execute(f"SET search_path = {search_path}")
            except (DatabaseError, psycopg2.InternalError):
                self._search_path_set = False
            else:
                self._search_path_set = True
            if name:
                cursor_for_search_path.close()
        return cursor



import argparse
import sys

from django.core.management import get_commands, load_command_class
from django.core.management.base import BaseCommand, CommandError, SystemCheckError

from . import WrappedSchemaOption


class Command(WrappedSchemaOption, BaseCommand):
    help = "Wrapper around Django commands for use with an individual schema"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("command_name", help="The command name you want to run")

    def get_command_from_arg(self, arg):
        *chunks, command = arg.split(".")
        path = ".".join(chunks)
        if not path:
            path = get_commands().get(command)
        try:
            cmd = load_command_class(path, command)
        except Exception:
            raise CommandError("Unknown command: %s" % arg)
        if isinstance(cmd, WrappedSchemaOption):
            raise CommandError("Command '%s' cannot be used in runschema" % arg)
        return cmd

    def run_from_argv(self, argv):  # pragma: no cover
        """
        Changes the option_list to use the options from the wrapped command.
        Adds schema parameter to specify which schema will be used when
        executing the wrapped command.
        """
        try:
            # load the command object.
            if len(argv) <= 2:
                raise CommandError("No command to run")
            target_class = self.get_command_from_arg(argv[2])
            # Ugly, but works. Delete command_name from the argv, parse the schemas manually
            # and forward the rest of the arguments to the actual command being wrapped.
            del argv[1]
            schema_parser = argparse.ArgumentParser()
            super().add_arguments(schema_parser)
            schema_ns, args = schema_parser.parse_known_args(argv)

            schemas = self.get_schemas_from_options(
                schemas=schema_ns.schemas,
                all_schemas=schema_ns.all_schemas,
                static_schemas=schema_ns.static_schemas,
                dynamic_schemas=schema_ns.dynamic_schemas,
                tenant_schemas=schema_ns.tenant_schemas,
            )
            executor = self.get_executor_from_options(parallel=schema_ns.parallel)
        except Exception as e:
            if not isinstance(e, CommandError):
                raise
            # SystemCheckError takes care of its own formatting.
            if isinstance(e, SystemCheckError):
                self.stderr.write(str(e), lambda x: x)
            else:
                self.stderr.write("%s: %s" % (e.__class__.__name__, e))
            sys.exit(1)

        executor(schemas, target_class, "special:run_from_argv", args)

    def handle(self, *args, **options):
        target = self.get_command_from_arg(options.pop("command_name"))
        schemas = self.get_schemas_from_options(**options)
        executor = self.get_executor_from_options(**options)
        options.pop("schemas")
        options.pop("excluded_schemas")
        options.pop("all_schemas")
        options.pop("static_schemas")
        options.pop("dynamic_schemas")
        options.pop("tenant_schemas")
        options.pop("parallel")
        options.pop("skip_schema_creation")
        if self.allow_interactive:
            options.pop("interactive")
        executor(schemas, target, "special:call_command", args, options)



import os

from django.conf import settings
from django.core.files.storage import FileSystemStorage

from ...schema import get_current_schema


class TenantFileSystemStorage(FileSystemStorage):
    """
    Tenant aware file system storage. Appends the tenant identifier to the base
    location and base URL.
    """

    def get_schema_path_identifier(self):
        current_schema = get_current_schema()
        if current_schema is None:
            return ""
        path_identifier = current_schema.schema_name
        if hasattr(current_schema, "schema_pathname"):
            path_identifier = current_schema.schema_pathname()
        elif hasattr(settings, "PGSCHEMAS_PATHNAME_FUNCTION"):
            path_identifier = settings.PGSCHEMAS_PATHNAME_FUNCTION(current_schema)
        return path_identifier

    @property  # To avoid caching of tenant
    def base_location(self):
        """
        Appends base location with the schema path identifier.
        """
        file_folder = self.get_schema_path_identifier()
        location = os.path.join(super().base_location, file_folder)
        if not location.endswith("/"):
            location += "/"
        return location

    @property  # To avoid caching of tenant
    def location(self):
        return super().location

    @property  # To avoid caching of tenant
    def base_url(self):
        """
        Optionally appends base URL with the schema path identifier.
        If the current schema is already using a folder, no path identifier is
        appended.
        """
        current_schema = get_current_schema()
        url_folder = self.get_schema_path_identifier()
        if url_folder and current_schema and current_schema.folder:
            # Since we're already prepending all URLs with schema, there is no
            # need to make the differentiation here
            url_folder = ""
        parent_base_url = super().base_url.strip("/")
        url = "/".join(["", parent_base_url, url_folder])
        if not url.endswith("/"):
            url += "/"
        return url



from channels.db import database_sync_to_async
from channels.routing import ProtocolTypeRouter, URLRouter
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import force_str
from django.utils.module_loading import import_string

from ...schema import SchemaDescriptor
from ...utils import get_domain_model, remove_www
from .auth import TenantAuthMiddlewareStack


class TenantAwareProtocolTypeRouter(ProtocolTypeRouter):
    def __init__(self, application_mapping, tenant_prefix):
        self.tenant_prefix = tenant_prefix
        super().__init__(application_mapping)

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            scope["path"] = scope["path"][len(self.tenant_prefix) + 1 :]
        return await super().__call__(scope, receive, send)


class TenantProtocolRouter:
    """
    ProtocolRouter that handles multi-tenancy.
    """

    def __init__(self):
        self.root_ws_urlconf = settings.TENANTS["default"].get("WS_URLCONF")
        if self.root_ws_urlconf is None:
            raise ImproperlyConfigured(
                "TENANTS['default'] must contain a 'WS_URLCONF' key in order to use TenantProtocolRouter."
            )

    @database_sync_to_async
    def get_tenant_scope(self, scope):
        """
        Get tenant and websockets urlconf based on scope host.
        """
        hostname = force_str(dict(scope["headers"]).get(b"host", b""))
        hostname = remove_www(hostname.split(":")[0])

        tenant = None
        ws_urlconf = self.root_ws_urlconf

        # Checking for static tenants
        for schema, data in settings.TENANTS.items():
            if schema in ["public", "default"]:
                continue
            if hostname in data["DOMAINS"]:
                tenant = SchemaDescriptor.create(schema_name=schema, domain_url=hostname)
                if "WS_URLCONF" in data:
                    ws_urlconf = data["WS_URLCONF"]
                return tenant, "", import_string(ws_urlconf + ".urlpatterns")

        # Checking for dynamic tenants
        else:
            DomainModel = get_domain_model()
            prefix = scope["path"].split("/")[1]
            try:
                domain = DomainModel.objects.select_related("tenant").get(domain=hostname, folder=prefix)
            except DomainModel.DoesNotExist:
                try:
                    domain = DomainModel.objects.select_related("tenant").get(domain=hostname, folder="")
                except DomainModel.DoesNotExist:
                    return None, "", []
            tenant = domain.tenant
            tenant.domain_url = hostname
            ws_urlconf = settings.TENANTS["default"]["WS_URLCONF"]
            return tenant, prefix if prefix == domain.folder else "", import_string(ws_urlconf + ".urlpatterns")

    def get_protocol_type_router(self, tenant_prefix, ws_urlconf):
        """
        Subclasses can override this to include more protocols.
        """
        return TenantAwareProtocolTypeRouter(
            {"websocket": TenantAuthMiddlewareStack(URLRouter(ws_urlconf))}, tenant_prefix
        )

    async def __call__(self, scope, receive, send):
        tenant, tenant_prefix, ws_urlconf = await self.get_tenant_scope(scope)
        scope.update({"tenant": tenant})
        return await self.get_protocol_type_router(tenant_prefix, ws_urlconf)(scope, receive, send)



from django.core.checks import Tags, run_checks
from django.core.management.base import BaseCommand
from django.core.management.commands.migrate import Command as MigrateCommand

from . import WrappedSchemaOption
from .runschema import Command as RunSchemaCommand


class NonInteractiveRunSchemaCommand(RunSchemaCommand):
    allow_interactive = False


class MigrateSchemaCommand(WrappedSchemaOption, BaseCommand):
    allow_interactive = False
    requires_system_checks = []

    def _run_checks(self, **kwargs):  # pragma: no cover
        issues = run_checks(tags=[Tags.database])
        issues.extend(super()._run_checks(**kwargs))
        return issues

    def add_arguments(self, parser):
        super().add_arguments(parser)
        MigrateCommand.add_arguments(self, parser)

    def handle(self, *args, **options):
        runschema = NonInteractiveRunSchemaCommand()
        options.pop("run_syncdb", False)
        runschema.execute(command_name="django.core.migrate", *args, **options)


Command = MigrateSchemaCommand



from django.conf import settings
from django.db import models, transaction

from .postgresql_backend.base import check_schema_name
from .schema import SchemaDescriptor
from .signals import dynamic_tenant_needs_sync, dynamic_tenant_post_sync, dynamic_tenant_pre_drop
from .utils import create_or_clone_schema, drop_schema, get_domain_model, schema_exists


class TenantMixin(SchemaDescriptor, models.Model):
    """
    All tenant models must inherit this class.
    """

    auto_create_schema = True
    """
    Set this flag to ``False`` on a parent class if you don't want the schema
    to be automatically created upon save.
    """

    auto_drop_schema = False
    """
    **USE THIS WITH CAUTION!**
    Set this flag to ``True`` on a parent class if you want the schema to be
    automatically deleted if the tenant row gets deleted.
    """

    is_dynamic = True
    """
    Leave this as ``True``. Denotes it's a database controlled tenant.
    """

    schema_name = models.CharField(max_length=63, unique=True, validators=[check_schema_name])

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        verbosity = kwargs.pop("verbosity", 1)
        is_new = self.pk is None

        super().save(*args, **kwargs)

        if is_new and self.auto_create_schema:
            try:
                self.create_schema(verbosity=verbosity)
                dynamic_tenant_post_sync.send(sender=TenantMixin, tenant=self.serializable_fields())
            except Exception:
                # We failed creating the tenant, delete what we created and re-raise the exception
                self.delete(force_drop=True)
                raise
        elif is_new:
            # Although we are not using the schema functions directly, the signal might be registered by a listener
            dynamic_tenant_needs_sync.send(sender=TenantMixin, tenant=self.serializable_fields())
        elif not is_new and self.auto_create_schema and not schema_exists(self.schema_name):
            # Create schemas for existing models, deleting only the schema on failure
            try:
                self.create_schema(verbosity=verbosity)
                dynamic_tenant_post_sync.send(sender=TenantMixin, tenant=self.serializable_fields())
            except Exception:
                # We failed creating the schema, delete what we created and re-raise the exception
                self.drop_schema()
                raise

    def delete(self, force_drop=False, *args, **kwargs):
        """
        Deletes this row. Drops the tenant's schema if the attribute
        ``auto_drop_schema`` is ``True``.
        """
        if force_drop or self.auto_drop_schema:
            dynamic_tenant_pre_drop.send(sender=TenantMixin, tenant=self.serializable_fields())
            self.drop_schema()
        super().delete(*args, **kwargs)

    def serializable_fields(self):
        """
        In certain cases the user model isn't serializable so you may want to
        only send the id.
        """
        return self

    def create_schema(self, sync_schema=True, verbosity=1):
        """
        Creates or clones the schema ``schema_name`` for this tenant.
        """
        return create_or_clone_schema(self.schema_name, sync_schema, verbosity)

    def drop_schema(self):
        """
        Drops the schema.
        """
        return drop_schema(self.schema_name)

    def get_primary_domain(self):
        try:
            domain = self.domains.get(is_primary=True)
            return domain
        except get_domain_model().DoesNotExist:
            return None


class DomainMixin(models.Model):
    """
    All models that store the domains must inherit this class.
    """

    tenant = models.ForeignKey(
        settings.TENANTS["default"]["TENANT_MODEL"], db_index=True, related_name="domains", on_delete=models.CASCADE
    )

    domain = models.CharField(max_length=253, db_index=True)
    folder = models.SlugField(max_length=253, blank=True, db_index=True)

    is_primary = models.BooleanField(default=True)
    redirect_to_primary = models.BooleanField(default=False)

    class Meta:
        abstract = True
        unique_together = (("domain", "folder"),)

    def __str__(self):
        return "/".join([self.domain, self.folder]) if self.folder else self.domain

    @transaction.atomic
    def save(self, *args, **kwargs):
        using = kwargs.get("using")
        domain_list = self.__class__.objects
        if using:
            domain_list = domain_list.using(using)
        domain_list = domain_list.filter(tenant=self.tenant, is_primary=True).exclude(pk=self.pk)
        self.is_primary = self.is_primary or (not domain_list.exists())
        if self.is_primary:
            domain_list.update(is_primary=False)
            if self.redirect_to_primary:
                self.redirect_to_primary = False
        super().save(*args, **kwargs)

    def absolute_url(self, path):
        """
        Constructs an absolute url for this domain / folder and a given path
        """
        folder = self.folder and "/" + self.folder
        if not path.startswith("/"):
            path = "/" + path
        return "//" + self.domain + folder + path



from django.db import connection

from .signals import schema_activate


def get_current_schema():
    return connection._schema


def activate(schema):
    connection._set_schema(schema)
    schema_activate.send(sender=SchemaDescriptor, schema=schema)


def deactivate():
    connection._set_schema_to_public()
    schema_activate.send(sender=SchemaDescriptor, schema=SchemaDescriptor.create("public"))


activate_public = deactivate


class SchemaDescriptor:
    schema_name = None
    domain_url = None
    folder = None

    is_dynamic = False

    @staticmethod
    def create(schema_name, domain_url=None, folder=None):
        tenant = SchemaDescriptor()
        tenant.schema_name = schema_name
        tenant.domain_url = domain_url
        tenant.folder = folder
        return tenant

    def __enter__(self):
        self.previous_schema = connection._schema
        activate(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        previous_schema = getattr(self, "previous_schema", None)
        activate(previous_schema) if previous_schema else deactivate()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_pgschemas',
 'django_pgschemas.contrib',
 'django_pgschemas.contrib.channels2',
 'django_pgschemas.contrib.channels3',
 'django_pgschemas.contrib.files',
 'django_pgschemas.management',
 'django_pgschemas.management.commands',
 'django_pgschemas.postgresql_backend',
 'django_pgschemas.test']

package_data = \
{'': ['*']}

install_requires = \
['django>=3.1,<4.2', 'psycopg2>=2.9.3,<3.0.0']

setup_kwargs = {
    'name': 'django-pgschemas',
    'version': '0.10.0',
    'description': 'Multi-tenancy on Django using PostgreSQL schemas.',
    'long_description': 'django-pgschemas\n================\n\n.. image:: https://img.shields.io/badge/packaging-poetry-purple.svg\n    :alt: Packaging: poetry\n    :target: https://github.com/sdispater/poetry\n\n.. image:: https://img.shields.io/badge/code%20style-black-black.svg\n    :alt: Code style: black\n    :target: https://github.com/ambv/black\n\n.. image:: https://github.com/lorinkoz/django-pgschemas/workflows/code/badge.svg\n    :alt: Build status\n    :target: https://github.com/lorinkoz/django-pgschemas/actions\n\n.. image:: https://readthedocs.org/projects/django-pgschemas/badge/?version=latest\n    :alt: Documentation status\n    :target: https://django-pgschemas.readthedocs.io/\n\n.. image:: https://coveralls.io/repos/github/lorinkoz/django-pgschemas/badge.svg?branch=master\n    :alt: Code coverage\n    :target: https://coveralls.io/github/lorinkoz/django-pgschemas?branch=master\n\n.. image:: https://badge.fury.io/py/django-pgschemas.svg\n    :alt: PyPi version\n    :target: http://badge.fury.io/py/django-pgschemas\n\n.. image:: https://pepy.tech/badge/django-pgschemas/month\n    :alt: Downloads\n    :target: https://pepy.tech/project/django-pgschemas/\n\n|\n\nThis app uses PostgreSQL schemas to support data multi-tenancy in a single\nDjango project. It is a fork of `django-tenants`_ with some conceptual changes:\n\n- There are static tenants and dynamic tenants. Static tenants can have their\n  own apps and urlconf.\n- Tenants can be simultaneously routed via subdomain and via subfolder on shared\n  subdomain.\n- Public is no longer the schema for storing the main site data. Public should\n  be used only for true shared data across all tenants. Table "overriding" via\n  search path is no longer encouraged.\n- Management commands can be run on multiple schemas via wildcards - the\n  multiproc behavior of migrations was extended to just any tenant command.\n\n.. _django-tenants: https://github.com/tomturner/django-tenants\n\nWhich package to use?\n---------------------\n\nThere are currently multiple packages to handle multi-tenancy via PostgreSQL schemas.\nThis table should help you make an informed decision on which one to choose.\n\n.. list-table::\n   :widths: 50 50\n   :header-rows: 1\n\n   * - Package\n     - Features\n   * - `django-tenant-schemas`_\n     - Original project.\n       Active and maintained by `@goodtune`_.\n   * - `django-tenants`_\n     - Active and maintained by `@tomturner`_.\n       Built on top of `django-tenant-schemas`_.\n       Uses a ``Domain`` model for allowing multiple domains per tenant.\n       Allows for parallel migrations with custom migration executor.\n       Other multiple improvements.\n   * - `django-pgschemas`_\n     - Active and maintained by `@lorinkoz`_.\n       Built on top of `django-tenants`_.\n       Different philosphy for tenants.\n       Other improvements listed above.\n\n.. _django-tenants-schemas: https://github.com/bernardopires/django-tenant-schemas\n.. _@goodtune: https://github.com/goodtune\n.. _django-tenants: https://github.com/tomturner/django-tenants\n.. _@tomturner: https://github.com/tomturner\n.. _django-pgschemas: https://github.com/lorinkoz/django-pgschemas\n.. _@lorinkoz: https://github.com/lorinkoz\n\nDocumentation\n-------------\n\nhttps://django-pgschemas.readthedocs.io/\n\nBreaking changes\n----------------\n\nv0.9.0\n++++++\n\n- Dropped support for Python < 3.8, Django < 3.1.\n\nv0.7.0\n++++++\n\n- Changed public API for getting/setting active schema. Public API is now\n  ``get_current_schema``, ``activate(schema)``, ``activate_public()``. Any\n  schema descriptor can still be used as context manager.\n- Changed location of tenant model and domain model in settings.\n  ``TENANT_MODEL`` and ``DOMAIN_MODEL`` keys are now under ``TENANTS["default"]``\n  instead of ``TENANTS["public"]``. This is required for future\n  static-tenant-only configurations.\n- Module ``cache`` renamed to ``contrib.cache``.\n- Module ``contrib.channels`` renamed to ``contrib.channels2``.\n- Added module ``contrib.channels3``.\n- Management command option ``--executor {sequential, parallel}`` renamed to\n  ``--parallel``.\n- All signals renamed. Added ``schema_activate`` signal.\n\nContributing\n------------\n\n- Join the discussion at https://github.com/lorinkoz/django-pgschemas/discussions.\n- PRs are welcome! If you have questions or comments, please use the link\n  above.\n- To run the test suite run ``make`` or ``make coverage``. The tests for this\n  project live inside a small django project called ``dpgs_sandbox``. Database\n  password and database host can be set through the environment variables\n  ``DATABASE_PASSWORD`` and ``DATABASE_HOST``.\n\nCredits\n-------\n\n* Tom Turner for `django-tenants`_\n* Bernardo Pires for `django-tenant-schemas`_\n\n.. _django-tenants: https://github.com/tomturner/django-tenants\n.. _django-tenant-schemas: https://github.com/bernardopires/django-tenant-schemas\n',
    'author': 'Lorenzo Peña',
    'author_email': 'lorinkoz@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lorinkoz/django-pgschemas',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from distutils.util import strtobool

from django.core.checks import Tags, run_checks
from django.core.management.base import BaseCommand, CommandError

from ...utils import clone_schema, get_domain_model, get_tenant_model


class Command(BaseCommand):
    help = "Clones a schema"

    def _run_checks(self, **kwargs):  # pragma: no cover
        issues = run_checks(tags=[Tags.database])
        issues.extend(super()._run_checks(**kwargs))
        return issues

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("source", help="The name of the schema you want to clone")
        parser.add_argument("destination", help="The name of the schema you want to create as clone")
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            help="Tells Django to NOT prompt the user for input of any kind.",
        )
        parser.add_argument(
            "--dry-run",
            dest="dry_run",
            action="store_true",
            help="Just show what clone would do; without actually cloning.",
        )

    def _ask(self, question):
        answer = None
        while answer is None:
            try:
                raw_answer = input(f"{question.strip()} [Y/n] ").strip() or "y"
                answer = strtobool(raw_answer)
            except ValueError:
                self.stderr.write(f"{raw_answer} is not a valid answer.")
                pass
        return answer

    def _check_required_field(self, field, exclude=None):
        if exclude is None:
            exclude = []
        return (
            field.editable
            and not field.primary_key
            and not field.is_relation
            and not (
                field.null
                or field.has_default()
                or (field.blank and field.empty_strings_allowed)
                or getattr(field, "auto_now", False)
                or getattr(field, "auto_now_add", False)
            )
            and field.name not in exclude
        )

    def _get_constructed_instance(self, model_class, data):
        fields = [field for field in model_class._meta.fields if self._check_required_field(field, data.keys())]
        instance = model_class(**data)
        if fields:
            self.stdout.write(self.style.WARNING(f"We need some data for model '{model_class._meta.model_name}':"))
            for field in fields:
                while field.name not in data:
                    raw_value = input(f"Value for field '{field.name}': ")
                    try:
                        data[field.name] = field.clean(raw_value, None)
                        instance = model_class(**data)
                        instance.clean()
                    except Exception as e:
                        if hasattr(e, "message"):
                            self.stderr.write(e.message)  # noqa
                        elif hasattr(e, "messages"):
                            self.stderr.write(" ".join(e.messages))  # noqa
                        else:
                            self.stderr.write(e)
                        data.pop(field.name, None)
        return instance

    def get_dynamic_tenant(self, **options):
        tenant = None
        domain = None
        if self._ask(
            "You are cloning a schema for a dynamic tenant. Would you like to create a database entry for it?"
        ):
            tenant = self._get_constructed_instance(get_tenant_model(), {"schema_name": options["destination"]})
            domain = self._get_constructed_instance(get_domain_model(), {"is_primary": True})
            if options["verbosity"] >= 1:
                self.stdout.write(self.style.WARNING("Looks good! Let's get to it!"))
        return tenant, domain

    def handle(self, *args, **options):
        tenant = None
        domain = None
        dry_run = options.get("dry_run")
        if options.get("interactive", True):
            TenantModel = get_tenant_model()
            if TenantModel.objects.filter(schema_name=options["source"]).exists():
                tenant, domain = self.get_dynamic_tenant(**options)
        try:
            clone_schema(options["source"], options["destination"], dry_run)
            if tenant and domain:
                if options["verbosity"] >= 1:
                    self.stdout.write("Schema cloned.")
                if not dry_run:
                    tenant.save()
                domain.tenant = tenant
                if not dry_run:
                    domain.save()
                if options["verbosity"] >= 1:
                    self.stdout.write("Tenant and domain successfully saved.")
            if options["verbosity"] >= 1:
                self.stdout.write("All done!")
        except Exception as e:
            if hasattr(e, "message"):
                raise CommandError(e.message)  # noqa
            elif hasattr(e, "messages"):
                raise CommandError(" ".join(e.messages))  # noqa
            else:
                raise CommandError(e)



import logging

from .schema import get_current_schema


class SchemaContextFilter(logging.Filter):
    """
    Add the current ``schema_name`` and ``domain_url`` to log records.
    """

    def filter(self, record):
        current_schema = get_current_schema()
        record.schema_name = current_schema.schema_name
        record.domain_url = current_schema.domain_url
        return True



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import connection

from .utils import get_tenant_model, is_valid_schema_name


class DjangoPGSchemasConfig(AppConfig):
    name = "django_pgschemas"
    verbose_name = "Django PostgreSQL Schemas"

    def _check_tenant_dict(self):
        if not isinstance(getattr(settings, "TENANTS", None), dict):
            raise ImproperlyConfigured("TENANTS dict setting not set.")

    def _check_public_schema(self):
        if not isinstance(settings.TENANTS.get("public"), dict):
            raise ImproperlyConfigured("TENANTS must contain a 'public' dict.")
        if "URLCONF" in settings.TENANTS["public"]:
            raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'URLCONF' key.")
        if "WS_URLCONF" in settings.TENANTS["public"]:
            raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'WS_URLCONF' key.")
        if "DOMAINS" in settings.TENANTS["public"]:
            raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'DOMAINS' key.")
        if "FALLBACK_DOMAINS" in settings.TENANTS["public"]:
            raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'FALLBACK_DOMAINS' key.")

    def _check_default_schemas(self):
        if not isinstance(settings.TENANTS.get("default"), dict):
            raise ImproperlyConfigured("TENANTS must contain a 'default' dict.")
        if "TENANT_MODEL" not in settings.TENANTS["default"]:
            raise ImproperlyConfigured("TENANTS['default'] must contain a 'TENANT_MODEL' key.")
        if "DOMAIN_MODEL" not in settings.TENANTS["default"]:
            raise ImproperlyConfigured("TENANTS['default'] must contain a 'DOMAIN_MODEL' key.")
        if "URLCONF" not in settings.TENANTS["default"]:
            raise ImproperlyConfigured("TENANTS['default'] must contain a 'URLCONF' key.")
        if "DOMAINS" in settings.TENANTS["default"]:
            raise ImproperlyConfigured("TENANTS['default'] cannot contain a 'DOMAINS' key.")
        if "FALLBACK_DOMAINS" in settings.TENANTS["default"]:
            raise ImproperlyConfigured("TENANTS['default'] cannot contain a 'FALLBACK_DOMAINS' key.")
        if (
            "CLONE_REFERENCE" in settings.TENANTS["default"]
            and settings.TENANTS["default"]["CLONE_REFERENCE"] in settings.TENANTS
        ):
            raise ImproperlyConfigured("TENANTS['default']['CLONE_REFERENCE'] must be a unique schema name.")

    def _check_overall_schemas(self):
        for schema in settings.TENANTS:
            if schema not in ["public", "default"]:
                if not is_valid_schema_name(schema):
                    raise ImproperlyConfigured("'%s' is not a valid schema name." % schema)
                if not isinstance(settings.TENANTS[schema].get("DOMAINS"), list):
                    raise ImproperlyConfigured("TENANTS['%s'] must contain a 'DOMAINS' list." % schema)

    def _check_complementary_settings(self):
        if "django_pgschemas.routers.SyncRouter" not in settings.DATABASE_ROUTERS:
            raise ImproperlyConfigured("DATABASE_ROUTERS setting must contain 'django_pgschemas.routers.SyncRouter'.")

    def _check_extra_search_paths(self):
        if hasattr(settings, "PGSCHEMAS_EXTRA_SEARCH_PATHS"):
            TenantModel = get_tenant_model()
            cursor = connection.cursor()
            cursor.execute(
                "SELECT 1 FROM information_schema.tables WHERE table_name = %s;", [TenantModel._meta.db_table]
            )
            dynamic_tenants = []
            if "CLONE_REFERENCE" in settings.TENANTS["default"]:
                dynamic_tenants.append(settings.TENANTS["default"]["CLONE_REFERENCE"])
            if cursor.fetchone():
                dynamic_tenants += list(TenantModel.objects.all().values_list("schema_name", flat=True))
            cursor.close()
            invalid_schemas = set(settings.PGSCHEMAS_EXTRA_SEARCH_PATHS).intersection(
                set(settings.TENANTS.keys()).union(dynamic_tenants)
            )
            if invalid_schemas:
                raise ImproperlyConfigured(
                    "Do not include '%s' on PGSCHEMAS_EXTRA_SEARCH_PATHS." % ", ".join(invalid_schemas)
                )

    def ready(self):
        from . import checks  # noqa

        self._check_tenant_dict()
        self._check_public_schema()
        self._check_default_schemas()
        self._check_overall_schemas()
        self._check_complementary_settings()
        self._check_extra_search_paths()
